"use strict";
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __metadata = (this && this.__metadata) || function (k, v) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(k, v);
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ProcessProvider = void 0;
const common_1 = require("@nestjs/common");
const child_process_1 = require("child_process");
const path_1 = require("path");
const util_1 = __importDefault(require("util"));
const guidHelper_1 = require("../../helpers/guidHelper");
const symbolType_1 = require("../binary/model/symbolType");
const config_service_1 = require("../config/config.service");
const logger_service_1 = require("../logger/logger.service");
const metrics_service_1 = require("../metrics/metrics.service");
const directoryProvider_1 = require("./directoryProvider");
const fileProvider_1 = require("./fileProvider");
const exe = util_1.default.promisify(child_process_1.execFile);
let ProcessProvider = class ProcessProvider {
    constructor(_metrics, _logger, _config, _directoryProvider, _fileProvider) {
        this._metrics = _metrics;
        this._logger = _logger;
        this._config = _config;
        this._directoryProvider = _directoryProvider;
        this._fileProvider = _fileProvider;
    }
    async extractCabinetsFromSymbols(symbols, destination) {
        for (const symbol of symbols) {
            const symbolType = symbol.type;
            if (symbolType === symbolType_1.SymbolType.Cabinet) {
                const extractResult = await this.extractCabinet(symbol.symbolPath, destination);
                if (!extractResult) {
                    ++this._metrics.data.errorsInCabinetUnpacking;
                    ++symbol.task.symbolServer.statistics.numberOfFailuresAtSymbolConversionStage;
                    this._logger.warn(`Cannot extract cabinet file available in a path: ${symbol.symbolPath}`);
                    continue;
                }
                ++this._metrics.data.successInCabinetUnpacking;
                ++symbol.task.symbolServer.statistics.numberOfFailuresAtSymbolConversionStage;
            }
        }
    }
    async extractCabinet(source, destination) {
        const tempDir = path_1.join(path_1.dirname(source), guidHelper_1.Guid.NewGuid());
        await this._directoryProvider.createDir(tempDir);
        const command = `${this._config.path.cabExtractCommand}`;
        this._logger.debug(`Trying to extract cabinet with command ${command} -d ${tempDir} ${source}`);
        try {
            await exe(this._config.path.cabExtractCommand, ['-d', tempDir, source]);
            this._logger.debug(`Successfully extracted cabinet file. Destination: ${destination}`);
            return true;
        }
        catch (err) {
            try {
                await this._fileProvider.delete(source);
            }
            catch (innerErr) {
                this._logger.warn(`Cannot delete directory. Reason: ${innerErr}`);
            }
            return false;
        }
    }
    async dumpSyms(source, destination) {
        const command = `${this._config.path.dumpSyms} -b -R -o ${destination} ${source}`;
        this._logger.debug(`Trying to convert symbols with command ${command}`);
        const sourceDirPath = path_1.dirname(path_1.dirname(path_1.dirname(source)));
        try {
            const { stdout, stderr } = await exe(this._config.path.dumpSyms, ['-b', '-R', '-o', destination, source], {
                cwd: sourceDirPath,
            });
            if (stderr && !stderr.startsWith('Warning')) {
                this._logger.warn(`dump_syms warning/error message: ${stderr}.`);
            }
            else {
                this._logger.info(`dump_syms: Successfully converted symbols. Destination: ${destination}.`);
                if (stdout) {
                    this._logger.debug(`dump_syms result: ${stdout}`);
                }
            }
            return true;
        }
        catch (err) {
            this._logger.warn(`Cannot convert symbols available in path: ${source}. Command: ${command}. Exit code: ${err.code} Error message: ${err.message}`);
            return false;
        }
    }
    async pdbConvert(source, destination) {
        const command = `${this._config.path.pdbConvert} -b -r -s ${destination} ${source}`;
        this._logger.debug(`Trying to convert symbols with command ${command}`);
        const files = (await this._directoryProvider.listAllFilesInDir(source)).join(',');
        this._logger.info(`Available symbols in source directory: ${files}`);
        try {
            const { stdout, stderr } = await exe(this._config.path.pdbConvert, [
                '-b',
                '-r',
                '-s',
                destination,
                source,
            ]);
            if (stderr && !stderr.startsWith('Warning')) {
                this._logger.warn(`Pdb_convert warning/error message: ${stderr}.`);
            }
            else {
                ++this._metrics.data.successInPdbConvert;
                this._logger.info(`pdb_convert: Successfully converted symbols. Destination: ${destination}. Files in dir  ${files}`);
                if (stdout) {
                    this._logger.debug(`Pdb_convert result: ${stdout}`);
                }
            }
            return true;
        }
        catch (err) {
            ++this._metrics.data.errorsInPdbConvert;
            this._logger.warn(`Cannot convert symbols available in path: ${files}. Command: ${command}. Exit code: ${err.code} Error message: ${err.message}`);
            return false;
        }
    }
    async zstdDecompress(src, dst) {
        try {
            await exe(this._config.path.zstd, ['-d', '-o', dst, src]);
            await this._directoryProvider.copyStat(src, dst);
            await this._directoryProvider.remove(src);
            ++this._metrics.data.successInZstdDecompress;
        }
        catch (err) {
            this._logger.warn(`${src}: decompression failed: ${err}`);
            ++this._metrics.data.errorsInZstdDecompress;
            return false;
        }
        return true;
    }
};
ProcessProvider = __decorate([
    common_1.Injectable(),
    common_1.Global(),
    __metadata("design:paramtypes", [metrics_service_1.MetricsService,
        logger_service_1.LoggerService,
        config_service_1.ConfigService,
        directoryProvider_1.DirectoryProvider,
        fileProvider_1.FileProvider])
], ProcessProvider);
exports.ProcessProvider = ProcessProvider;
//# sourceMappingURL=processProvider.js.map