"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.BacktraceService = void 0;
var argparse_1 = require("argparse");
var body_parser_1 = require("body-parser");
var compression_1 = __importDefault(require("compression"));
var cors_1 = __importDefault(require("cors"));
var express_1 = __importDefault(require("express"));
var helmet_1 = __importDefault(require("helmet"));
var nocache_1 = __importDefault(require("nocache"));
var __1 = require("..");
var logger_1 = require("../log/logger");
var metricsStorage_1 = require("../metrics/metricsStorage");
var projectValidation_1 = require("./projectValidation");
var serviceConfiguration_1 = require("./serviceConfiguration");
// tslint:disable-next-line: no-var-requires
var packageJson = require('../../package.json');
var BacktraceService = /** @class */ (function () {
    function BacktraceService(name, port, _opts) {
        var _a, _b, _c, _d, _e, _f, _g, _h, _j;
        this.name = name;
        this.port = port;
        this._opts = _opts;
        this.coronerdAddress = new Set();
        this.app = (_opts === null || _opts === void 0 ? void 0 : _opts.app) ? _opts.app : express_1.default();
        this.initializeDefaultArgParser((_a = this._opts) === null || _a === void 0 ? void 0 : _a.serviceMutation);
        this.configuration = BacktraceService.readConfiguration(this.name, this.port, ((_b = this.arguments) === null || _b === void 0 ? void 0 : _b.configuration) || ((_c = this._opts) === null || _c === void 0 ? void 0 : _c.confPath), false);
        this.apiPrefix = this.configuration.descriptor
            ? this.configuration.descriptor.resource
            : "api/" + name;
        var logLevel = this.configuration.logger && this.configuration.logger.logLevel
            ? this.configuration.logger.logLevel
            : undefined;
        this._logger =
            this._opts && this._opts.logger
                ? this._opts.logger
                : new logger_1.Logger(logLevel);
        if (this.isBacktraceIntegrationEnabled()) {
            logger_1.Logger.initializeBacktrace(this.configuration.backtrace);
        }
        /**
         * In previous version of backtrace-service we assumed that MetricsStorage
         * will always exists. To make sure metricsStorage object is always available
         * for services that already use this object from BacktraceService
         * we will pass empty uuid/secret if remoteLog is disabled.
         */
        var _k = this.getMetricsSecret(), uuid = _k.uuid, secret = _k.secret;
        this.metrics = new metricsStorage_1.MetricsStorage(uuid, secret, {
            interval: (_e = (_d = this.configuration) === null || _d === void 0 ? void 0 : _d.metrics) === null || _e === void 0 ? void 0 : _e.interval,
            hostUrl: (_g = (_f = this.configuration) === null || _f === void 0 ? void 0 : _f.metrics) === null || _g === void 0 ? void 0 : _g.hostUrl,
            disableExpressMetrics: (_j = (_h = this.configuration) === null || _h === void 0 ? void 0 : _h.metrics) === null || _j === void 0 ? void 0 : _j.disableExpressMetrics,
        }, this.app, this._logger);
        this.prepareServiceLayer();
        BacktraceService._service = this;
    }
    Object.defineProperty(BacktraceService.prototype, "arguments", {
        /**
         * Returns ArgumentParser instance arguments.
         */
        get: function () {
            if (!this._arguments) {
                this._arguments = this._argumentParser.parseKnownArgs()[0];
            }
            return this._arguments;
        },
        enumerable: false,
        configurable: true
    });
    /**
     * Create and startup express application with basic options.
     * By using this method, application will apply all
     * Backtrace service rules
     * @param name service name (required by service layer)
     * @param port service port (required by service layer)
     * @param opts optional application options
     * @param expressApp - deprecated. Optional express app instance.
     * Please use expressInstance from app instead.
     * By using this parameter you can create express app in different
     * framework and still be able to use Backtrace integrations
     */
    BacktraceService.bootstrap = function (name, port, opts, app) {
        if (!port || !name) {
            throw new Error("Invalid arguments. Name: " + name + ", port: " + port);
        }
        // backward compatibility
        if (app && opts) {
            opts.app = app;
        }
        var service = new BacktraceService(name, port, opts);
        // assign express intsance prepared by backtrace service/passed by user
        var expressApp = service.app;
        this.setupAuth(expressApp);
        this.setupMiddlewares(expressApp, service.apiPrefix, service.identityManager);
        this.setupServiceLayer(expressApp, name, port, function (ip) {
            service.coronerdAddress.add(ip);
            if (opts && opts.coronerdCallback) {
                opts.coronerdCallback(ip);
            }
        });
        process.title = name;
        expressApp.on('close', function () {
            service._logger.warn('server is closing');
        });
        return service;
    };
    /**
     * Read and parse service configuration
     * @param name service name
     * @param port service port number
     * @param path (optional) service configuration path
     * @param ignoreDescriptor (optional) determine if service
     * should create/read service layer service descriptor
     * @param parser argParser instance to read command line arguments
     */
    BacktraceService.readConfiguration = function (name, port, path, ignoreDescriptor) {
        var serviceConfiguration = new serviceConfiguration_1.ServiceConfiguration();
        var possiblePaths = path ? [path] : [];
        possiblePaths.push.apply(possiblePaths, serviceConfiguration.getProjectSettingsPaths(name));
        var configuration = serviceConfiguration.loadConfiguration(possiblePaths);
        if (!ignoreDescriptor) {
            configuration.descriptor = __1.getDescriptor(name, port);
        }
        return configuration;
    };
    /**
     * Setup popular Backtrace middlewares
     * @param app express app
     */
    BacktraceService.setupMiddlewares = function (app, prefix, identityManager) {
        app.use(body_parser_1.json());
        app.use(body_parser_1.urlencoded({ extended: true }));
        app.use(compression_1.default());
        // Add default Backtrace headers to every service response
        app.use(function (req, res, next) {
            // add service version
            res.set('service-version', process.env.npm_package_version);
            // add backtrace-service version
            res.set('bt-service-version', packageJson.version);
            next();
        });
        // Route hit by pingdom
        app.get(prefix ? prefix : '/', function (_req, res) {
            identityManager === null || identityManager === void 0 ? void 0 : identityManager.bypassProjectAccessCheck(res);
            res.type('xml');
            res.status(200).send("<html></html>");
        });
    };
    /**
     * Setup express error middleware
     * @param app express app
     */
    BacktraceService.setupErrorMiddleware = function (app, identityManager) {
        app.use(function (err, req, res, next) {
            // handle a situation when we already sent a response from onError method
            // we shouldn't do that and we should let error middleware handle that
            // but here we wan't to make sure our code won't break up
            // and we won't generate another error
            if (res.headersSent) {
                return;
            }
            var logger = logger_1.Logger.getLogger();
            // Backtrace service ApiError (pending in CR) contains type statusCode that allows us
            // to determine what type of response we should return
            // in addition to that I recognize sometimes we retrieve status code from response.status
            //  property. This is a popular pattern for axios communication - check comments and hooks.
            if (err.statusCode || (err.response && err.response.status)) {
                var status_1 = err.statusCode || err.response.status;
                var message = err.message || err.response.message;
                if (status_1 === 500) {
                    logger.error(err);
                }
                else {
                    logger.warn(message);
                }
                res.status(status_1).send(message);
                return;
            }
            logger.error(err);
            res.status(500).send({
                error: 'Internal server error',
            });
        });
        app.use(function (req, res, next) {
            // prevent overriding headers.
            if (res.headersSent) {
                return;
            }
            identityManager === null || identityManager === void 0 ? void 0 : identityManager.bypassProjectAccessCheck(res);
            // response json might be required by front-end
            res.status(404).send({
                error: 'Not found',
            });
        });
    };
    /**
     * Setup security middlewares
     * @param app express app
     */
    BacktraceService.setupAuth = function (app) {
        var corsSettings = cors_1.default({
            exposedHeaders: ['service-version', 'bt-service-version'],
        });
        app.use(corsSettings);
        app.options('*', corsSettings);
        app.use(nocache_1.default());
        app.use(helmet_1.default({
            hsts: false,
        }));
        app.use(projectValidation_1.projectValidation);
    };
    /**
     * Setup service layer integration
     * @param app express app
     */
    BacktraceService.setupServiceLayer = function (app, name, port, coronerdCallback) {
        var serviceDescriptor = this._service
            ? this._service._descriptor
            : undefined;
        // this case handle a situation when we didn't call backtraceService constr
        // here we want to read service descriptor from hard drive - if descr exists
        // we're using here a service name and port instead of service descriptor
        // to avoid extra steps required by service to setup service layer.
        if (!serviceDescriptor && name && port) {
            serviceDescriptor = __1.getDescriptor(name, port);
        }
        if (!serviceDescriptor) {
            throw new Error('Descriptor is undefined');
        }
        var identityManager = this._service
            ? this._service.identityManager
            : new __1.IdentityManager(serviceDescriptor);
        app.post(serviceDescriptor.resource + "/service", identityManager.serviceRequest({
            name: serviceDescriptor.name,
            secret: serviceDescriptor.secret,
            coronerdCallback: coronerdCallback,
        }));
    };
    BacktraceService.prototype.start = function (clusterSupport) {
        var _this = this;
        if (clusterSupport === void 0) { clusterSupport = false; }
        if (!this.app) {
            this._logger.error("Express application is not defined. Cannot start application");
            return;
        }
        if (!this._descriptor) {
            this._logger.error("Descriptor object is not defined.\n        Please execute setupServiceLayer method");
            return;
        }
        BacktraceService.setupErrorMiddleware(this.app, this.identityManager);
        __1.listenDescriptor(this._descriptor, this.app, this.configuration.ssl, function () {
            _this._logger.info("Application is Running. " + (_this._descriptor ? "Port: " + _this._descriptor.port : ''));
        }, clusterSupport);
    };
    /**
     * Initialize default arg parser object.
     */
    BacktraceService.prototype.initializeDefaultArgParser = function (serviceMutation) {
        this._argumentParser = new argparse_1.ArgumentParser();
        this._argumentParser.addArgument(['-c', '--configuration'], {
            help: 'Override service configuration path',
        });
        if (serviceMutation) {
            serviceMutation(this);
        }
    };
    /**
     * Determine if backtrace-service should use backtrace-node integration
     */
    BacktraceService.prototype.isBacktraceIntegrationEnabled = function () {
        var _a;
        return ((_a = this.configuration.backtrace) === null || _a === void 0 ? void 0 : _a.enabled) === true;
    };
    /**
     * Generate correct credentials for metricsStorage obejct based on configuration file.
     */
    BacktraceService.prototype.getMetricsSecret = function () {
        var _a, _b, _c;
        var isDiagnosticsEnabled = ((_a = this.configuration.metrics) === null || _a === void 0 ? void 0 : _a.enabled) === true;
        if (isDiagnosticsEnabled === false) {
            return {
                uuid: '',
                secret: '',
            };
        }
        return {
            uuid: (_b = this.configuration.metrics) === null || _b === void 0 ? void 0 : _b.uuid,
            secret: (_c = this.configuration.metrics) === null || _c === void 0 ? void 0 : _c.secret,
        };
    };
    BacktraceService.prototype.prepareServiceLayer = function () {
        if (!this._descriptor) {
            var descriptor = __1.getDescriptor(this.name, this.port);
            this._descriptor = descriptor;
        }
        this.identityManager = new __1.IdentityManager(this._descriptor);
    };
    return BacktraceService;
}());
exports.BacktraceService = BacktraceService;
//# sourceMappingURL=backtraceService.js.map